#include "pdu.h"
#include <stdio.h>

//***********************************************************
// Generate reverse conversion table from SMS7Bit table.
//***********************************************************
GenerateReverseTable(unsigned char table[]){

int i;
extern unsigned char Sms7Bit[];

for(i=0; i<128; i++){
	table[Sms7Bit[i]]=i;
	}
}

/***********************************************************
MsgToTpdu function converting the plain ASCII message to hex-coded TPDU
(Transport PDU) string.
Parameters:
int vp_5 - SMS validity period in 5min intervals ( will be rounded to the
						closest available value according to TP_VP definition) .
int srr  - SMS status report request flag (0= no request, 1=request SRR)
int mr   - Integer representing the Message Reference value (0...255)
char *da - Destination Address max 20 digits in ASCII (usually a subscriber
					 number)
char *txt- SMS message text (max 160 ASCII characters)
char *s  - The resulting sms_submit structure

The function returns an integer result code depending on the successful
conversion:
SMS_OK      - Successful conversion
SMS_MR_ERR  - Invalid Message Reference value
SMS_DA_ERR  - Incorrect Destination Address format
SMS_TXT_ERR - SMS message text length exeeds 160 characters

************************************************************/

int MsgToTpdu(int vp_5,int srr,int mr,char da[],unsigned char txt[], sms_submit *s){

int i;
int da_length;
int da_octets;  // Number of octets in da field
int txt_length;
char da_[21];  // Copy of da without '+'

// variables for 7bit->8bit conversion
extern char Sms7BitRev[];
int bit_addr;  // Bit number in the pdu_ud field
int i_ch;     // Bit number in the character to be processed
int ch;


// Parameter check

if( (mr < 0) || (mr > 255) ){
	return SMS_MR_ERR; // Incorrect MR
	}

da_length=strlen(da);
if( (da_length < 1 ) || (da_length > 21) ){
	return SMS_DA_ERR; // Incorrect DA length
	}
for( i=1; i<da_length; i++ ){
	if( (da[i] < '0') || (da[i] > '9') ){
		return SMS_DA_ERR; // Incorrect DA format
		}
	}
txt_length=strlen(txt);
if( (txt_length < 1 ) || (txt_length > 160) ){
	return SMS_DA_ERR; // Incorrect message textlength
	}

// Start assembling of the PDU

// The first byte containig the PDU type information etc..
if( srr ){   // Status report request?
	s->tp_mt=MTI_SMS_SUBMIT | TP_VPF_RELATIVE | MTI_SRR;
	}
else{
	s->tp_mt=MTI_SMS_SUBMIT | TP_VPF_RELATIVE;
	}

// The message reference (mr)...
s->tp_mr=mr;
// The Destination Address (da)...

if(da[0] == '+'){
	da_length--;
	s->tp_da[0]=da_length;
	s->tp_da[1]=DA_TON_INTERNATIONAL | DA_NPI_ISDN; //International prefix .
	strcpy(da_,da+1);
	}
else{
	s->tp_da[1]=DA_TON_UNKNOWN | DA_NPI_ISDN; // Default setting, normally ok.
	s->tp_da[0]=da_length;
	strcpy(da_,da);
	}

// Assemble the subscriber number field...
da_[da_length]='?'; // Add '1111' which is needed if the number of digits is odd
da_octets=2+(da_length/2) + (da_length%2); // Calculate the number of octets in da field
for( i=0; i < da_octets; i++){
	s->tp_da[i+2] = ( (da_[(i*2)+1] - '0') << 4 ) | (da_[i*2] -'0'); // ASCII -> Binary
	}
// TP-PID field
s->tp_pid=PID_IMPLICIT; // Default setting (No interworking, SME to SME)

// TP-DCS field
s->tp_dcs=0x00;    // Default alphabet as defined in GSM 03.38

// TP-VP field
// Convert the vp_5 value to the closest possible value according to the
// relative VP format definition
if( vp_5 < 144 ){
	s->tp_vp=vp_5;
	}
else if( ( vp_5 >= 144 ) && ( vp_5 <= 288 ) ){  // 12h ... 24h
	s->tp_vp=143 + ( vp_5/12 );
	}
else if( ( vp_5 >= 289 ) && ( vp_5 <= 8640 ) ){  // 2 days ... 30 days
	s->tp_vp=166 + ( vp_5/288 );
	}
else if ( vp_5 >= 8640 ){  // More than 30 days (up to 31 weeks can be set)
//                            using unsigned int as vp_5 parameter
	s->tp_vp=192 + ( vp_5/2016 );
	}
s->tp_udl=txt_length;
// Convert the message text from ASCII to GSM alphabet and fill the TP_UD
// field (7bit -> 8bit)
for (i=0; i<txt_length; i++){
	txt[i]=Sms7BitRev[txt[i]];	// ASCII -> GSM conversion;
	}
for(i=0; i<140; i++) s->tp_ud[i]=0;  // Clean the output buffer
bit_addr=0;
i=0;
while( i < txt_length){
	ch = txt[i];   // get next 7-bit character
	for( i_ch=0; i_ch < 7; i_ch++){ // Scan all 7 bits
		if(ch & 1) set_bit(bit_addr, s->tp_ud); // Set bits which should be 1's
		bit_addr++; // Next bit position in the output buffer
		ch >>= 1;   // Shift next bit to the right position
		}
	i++; // Next character
	}

s->nb=(bit_addr >> 3) +1;  // Calculate & Store the actual number of octets
return SMS_OK;

}

//***********************************************************
// This is a function that helps to set a bit in the data array to 1
// where the addr is the sequential number of the bit in the array
//***********************************************************

set_bit(int addr,char data[]){
int pos,index,mask;

pos = addr&7;  // Calculate the bit number to be processed (0...7)
index = addr >> 3; // Calculate the index of the octet
mask = 1 << pos;  // Calculate the mask value
data[index] |= mask; // Set masked bit to 1
}

/************************************************************
AssemblePDU() function to fill the pdu_str string which is used as input
data in the AT+CMGS (Send SMS) command.
The input to the function is a sms_submit structure ( TPDU data) filled by
MsgToTpdu() function and the Service Centre Address (SCA) string.
************************************************************/

AssemblePDU(unsigned char pdu_str[],char sca[],sms_submit *s){
int sca_length;  // sca string length
unsigned char sca_da[13]; // Converted SCA for PDU
int i;
int pos;
int sca_octets; // SCA length in octets

// The SCA (service center address) conversion
sca_length=strlen(sca); // Calculate SCA length
sca_da[0]= sca_length/2 + 1; // Two digits per octet + 1 control byte
if( sca_length & 1 ) sca_da[0] ++; // Increase length byte by one if odd number of digits
sca_da[1]=DA_TON_INTERNATIONAL | DA_NPI_ISDN; // "International" prefix always required.

sca[sca_length]='?'; // Add '1111' which is needed if the number of digits is odd
sca_octets=(sca_length/2)+(sca_length%2) + 2;
for( i=0; i<sca_octets; i++){
	sca_da[i+2] = ( (sca[(i*2)+1] - '0') << 4 ) | (sca[i*2] -'0'); // ASCII -> Binary
	}
// Assemble the final PDU string
pos=0;
s->length=0;
sca[sca_length]=0; // Restore the original SCA
for(i=0; i<sca_octets; i++){     // SCA addres
	sprintf(pdu_str+pos,"%02X",sca_da[i]);
	pos+=2;
	}

sprintf(pdu_str+pos,"%02X",s->tp_mt);  // MTI... field
pos+=2; s->length++;
sprintf(pdu_str+pos,"%02X",s->tp_mr);  // Message Reference
pos+=2; s->length++;
for(i=0; i < (s->tp_da[0]/2)+(s->tp_da[0]%2)+2; i++){    // Destination Address
	sprintf(pdu_str+pos,"%02X",s->tp_da[i]);
	pos+=2; s->length++;
	}
sprintf(pdu_str+pos,"%02X",s->tp_pid);  // Protocol ID
pos+=2; s->length++;
sprintf(pdu_str+pos,"%02X",s->tp_dcs); // Data Coding Scheme
pos+=2; s->length++;
sprintf(pdu_str+pos,"%02X",s->tp_vp); // Validity Period
pos+=2; s->length++;
sprintf(pdu_str+pos,"%02X",s->tp_udl); // User Data Length
pos+=2; s->length++;

for(i=0; i < s->nb; i++){
	sprintf(pdu_str+pos,"%02X",s->tp_ud[i]);  // User Data
	pos+=2; s->length++;
	}
pdu_str[pos]=0; // Terminate string
}

/************************************************************
TPduToMsg() function to convert the raw TPDU string to sms_deliver
structure and decode address and text fields. Input to the function is a raw
TPDU hex string without SCA field.
The output is a filled tpdu structure. The message text converted to ASCII
is stored in the tpdu.txt string.
************************************************************/

TPduToMsg(char tpdu_str[],sms_deliver *tpdu){
int i,ii;
int pos=0;
int oa_length;
int ch;
int bit_addr;
int sflag; // If the message read from the module is sms_submit type
// (e.g. Stored Sent message) the message format differs, but we still
// want to read the message so let's use this flag to indicate the different
// format to the processing part of the function
extern unsigned char Sms7Bit[];

sscanf(tpdu_str+pos,"%2X",&(tpdu->tp_mt)); // Fill the tp_mt field
if( tpdu->tp_mt & MTI_SMS_SUBMIT ) sflag=1;
else sflag=0;
pos+=2;
if(sflag) pos+=2; // Skip the MR field if SMS_SUBMIT type;
sscanf(tpdu_str+pos,"%2X",&oa_length); // Get the Origin Address Length
for(i=0; i <= (oa_length/2) + (oa_length%2) + 1;i++){
	sscanf(tpdu_str+pos,"%02X",&(tpdu->tp_oa[i]));  // Store the tp_oa octets
	pos+=2;
	}
tpdu->oa_txt[0]=' ';
if( (tpdu->tp_oa[1] & 0xf0) == DA_TON_INTERNATIONAL ){ // International number
	tpdu->oa_txt[0]='+';
	}

ii=2;
for(i=1; i<=oa_length; i+=2){ // Convert Origin Address to ASCII text
	tpdu->oa_txt[i]= '0' + (tpdu->tp_oa[ii] & 0x0f);  // Convert to ASCII and swap digits
	tpdu->oa_txt[i+1]= '0' + (tpdu->tp_oa[ii++] >> 4);
	}
tpdu->oa_txt[oa_length+1]=0; // Terminate string
sscanf(tpdu_str+pos,"%2X",&(tpdu->tp_pid)); // Fill the tp_pid field
pos+=2;
sscanf(tpdu_str+pos,"%2X",&(tpdu->tp_pid)); // Fill the tp_dcs field
pos+=2;
if( sflag ){
	strcpy(tpdu->tp_sct,"Not Available");
	pos+=2;
	}
else{
	for(i=0; i<14; i+=2){ // Get the Service Centre Time stamp (SCT)
		tpdu->tp_sct[i]=tpdu_str[pos+1]; // Swap octets
		tpdu->tp_sct[i+1]=tpdu_str[pos];
		pos+=2;
		}
	tpdu->tp_sct[14]=0; // Terminate the string
	}
sscanf(tpdu_str+pos,"%2X",&(tpdu->tp_udl)); // Fill the tp_udl field
pos+=2;
i=0;
while( tpdu_str[pos] != 0) {
	sscanf(tpdu_str+pos,"%2X",&(tpdu->tp_ud[i++])); // Fill the tp_ud field
	pos+=2;
	}
// Convert the tp_ud to plain ASCII text

for(i=0; i<160; i++) tpdu->txt[i]=0;  // Clean the output buffer
bit_addr=0;
pos=0;
while( pos < tpdu->tp_udl){
	ch=0;
	for(i=0; i < 7; i++ ){
		ch >>= 1;
		if( get_bit( bit_addr++, tpdu->tp_ud )) ch |= 0x40;
		}
	tpdu->txt[pos++]=Sms7Bit[ch]; // Use SMS <-> ASCII conversion table
	}
}

//***********************************************************
// Bit mapping for 7bit to 8bit conversion
//***********************************************************

get_bit(int addr,char data[]){
int pos,index,mask;

pos = addr&7;  // Calculate the bit number to be processed (0...7)
index = addr >> 3; // Calculate the index of the octet
mask = 1 << pos;  // Calculate the mask value
if( (data[index] & mask) == mask ) return 1; // Bit=1
else return 0;
}

